import java.awt.*;
import processing.core.*;
import processing.video.*;
import blobDetection.*;
import controlP5.*;	
import imageadjuster.*;

public class SwarmHunger extends PApplet{
	//CAM:
	public Capture cam;
	public static PGraphics videoBuffer;
	public BlobDetection theBlobDetection;
	public static int frameWidth = 280;
	public static int frameHeight = 210;
	public static int frameX = 250;
	public static int frameY = 40;
	public PImage img;
	public boolean newFrame=false;
	public float threshold = 0.35f;
	public float brightness = 1.0f;
	public float contrast = 1.0f;
	//IMAGE MANIPULATION:
	public static ImageAdjuster adjust;
	//DRAWING:
	private int circleDiameter = 20;
	public static PGraphics pg;
	public boolean showBlobs = true;
	public boolean showEdges = false;
	//GUI:
	private ControlP5 controlP5;
	private boolean controlWindowVisible = true;
	private ControlWindow controlWindow;
	private ControlWindowCanvas extraCanvas;
	private Toggle blobToggle;
	private Toggle edgeToggle;
	
	public static void main(String args[]) {
		//PApplet.main(new String[] { "SwarmHunger" });
	    PApplet.main(new String[] { "--present", "SwarmHunger" });
	}
	
	public static PGraphics getCam(){
		return videoBuffer;
	}
	
	public void setup(){
		size(1024, 768);
		/*GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
	    GraphicsDevice gs = ge.getDefaultScreenDevice();
		Window win = new Window(frame);
		gs.setFullScreenWindow(frame);
        win.validate();*/
		
		
		/*
		 * Initialize Camera + video stuff
		 */
		img = new PImage(frameWidth,frameHeight); 
		pg = createGraphics(frameWidth,frameHeight,P2D);
		videoBuffer = createGraphics(width, height, P2D);
		CaptureThread runner = new CaptureThread(this, videoBuffer);
		runner.start();
		theBlobDetection = new BlobDetection(img.width, img.height);
		theBlobDetection.setPosDiscrimination(true);
		theBlobDetection.setThreshold(threshold); // will detect bright areas whose luminosity > 0.2f;
		
		adjust = new ImageAdjuster(this);
		
		/*
		 * Initialize controller window + GUI
		 */
		controlP5 = new ControlP5(this);
		controlWindow = controlP5.addControlWindow("controlP5window",100,100,600,300);
		Controller radiusSlider = controlP5.addSlider("circleDiameter",0,100,40,40,100,10);
		radiusSlider.setWindow(controlWindow);
		Controller sensitivitySlider = controlP5.addSlider("threshold",0,1,40,70,100,10);
		sensitivitySlider.setWindow(controlWindow);
		Controller brightnessSlider = controlP5.addSlider("brightness",0,2,40,100,100,10);
		brightnessSlider.setWindow(controlWindow);
		Controller contrastSlider = controlP5.addSlider("contrast", 0, 2, 40, 130, 100, 10);
		contrastSlider.setWindow(controlWindow);
		blobToggle = controlP5.addToggle("showBlobs", showBlobs, 40,170,20,20);
		blobToggle.setWindow(controlWindow);
		edgeToggle = controlP5.addToggle("showEdges", showEdges, 130,170,20,20);
		edgeToggle.setWindow(controlWindow);
		
		extraCanvas = new MyCanvas();
		extraCanvas.post(); 
		controlWindow.addCanvas(extraCanvas);

		smooth();
		background(255);
	}

	public void draw(){
		Blob tempBlob;
		for(int i=0;i<theBlobDetection.getBlobNb();i++){
			tempBlob = theBlobDetection.getBlob(i);
			fill(0);
			noStroke();
			ellipse(tempBlob.x*width, tempBlob.y*height, circleDiameter, circleDiameter);
		}
		
		if (newFrame){
			newFrame=false;
			pg.beginDraw();
			img.copy(videoBuffer, 0, 0, videoBuffer.width, videoBuffer.height, 
					0, 0, img.width, img.height);
			fastblur(img, 2);
			if(contrast != 1.0f){
				adjust.contrast(img,contrast);
			}
			if(brightness != 1.0f){
				adjust.contrast(img,brightness);
			}
			theBlobDetection.computeBlobs(img.pixels);
			pg.image(img,0,0);
			drawBlobsAndEdges(showBlobs,showEdges);
			pg.endDraw();
		}
	}
	
	
	/*
	 * Deal with spacebar - toggle control window visibility
	 */
	public void keyPressed(){
		if(key == ' '){
			if(controlWindowVisible){
				controlP5.window("controlP5window").hide();
				controlWindowVisible = false;
			}
			else{
				controlP5.window("controlP5window").show();
				controlWindowVisible = true;
				controlP5.window("controlP5window").component().requestFocus();
			}
		}
		else if(key == 'c' || key == 'C'){
			background(255);
		}
	}
	
	/*
	 * Draw the webcam image into the canvas element and put it in the new window
	 */
	void controlEvent(ControlEvent theEvent) {
		  if(theEvent.controller().name() == "threshold"){
			  theBlobDetection.setThreshold(threshold);
		  }
		  if(theEvent.controller().name() == "brightness"){
			  brightness = 2.0f - brightness;
		  }
	}

	/*
	 * BlobDetection
	 */
	public void drawBlobsAndEdges(boolean drawBlobs, boolean drawEdges){
		pg.noFill();
		Blob b;
		EdgeVertex eA,eB;
		for (int n=0 ; n<theBlobDetection.getBlobNb() ; n++)
		{
			b=theBlobDetection.getBlob(n);
			if (b!=null)
			{
				// Edges
				if (drawEdges)
				{
					pg.strokeWeight(1);
					pg.stroke(0,0,200);
					for (int m=0;m<b.getEdgeNb();m++)
					{
						eA = b.getEdgeVertexA(m);
						eB = b.getEdgeVertexB(m);
						if (eA !=null && eB !=null)
							pg.line(
								eA.x*frameWidth, eA.y*frameHeight, 
								eB.x*frameWidth, eB.y*frameHeight
								);
					}
				}

				// Blobs
				if (drawBlobs){
					pg.strokeWeight(1);
					pg.stroke(0,255,0);
					pg.rect(
						b.xMin*frameWidth, b.yMin*frameHeight,
						b.w*frameWidth , b.h*frameHeight
						);
				}

			}

	    }
	}

	//==================================================
	//Super Fast Blur v1.1
	//by Mario Klingemann 
	//<http://incubator.quasimondo.com>
	//==================================================
	void fastblur(PImage img,int radius)
	{
	if (radius<1){
	  return;
	}
	int w=img.width;
	int h=img.height;
	int wm=w-1;
	int hm=h-1;
	int wh=w*h;
	int div=radius+radius+1;
	int r[]=new int[wh];
	int g[]=new int[wh];
	int b[]=new int[wh];
	int rsum,gsum,bsum,x,y,i,p,p1,p2,yp,yi,yw;
	int vmin[] = new int[max(w,h)];
	int vmax[] = new int[max(w,h)];
	int[] pix=img.pixels;
	int dv[]=new int[256*div];
	for (i=0;i<256*div;i++){
	  dv[i]=(i/div);
	}

	yw=yi=0;

	for (y=0;y<h;y++){
	  rsum=gsum=bsum=0;
	  for(i=-radius;i<=radius;i++){
	    p=pix[yi+min(wm,max(i,0))];
	    rsum+=(p & 0xff0000)>>16;
	    gsum+=(p & 0x00ff00)>>8;
	    bsum+= p & 0x0000ff;
	  }
	  for (x=0;x<w;x++){

	    r[yi]=dv[rsum];
	    g[yi]=dv[gsum];
	    b[yi]=dv[bsum];

	    if(y==0){
	      vmin[x]=min(x+radius+1,wm);
	      vmax[x]=max(x-radius,0);
	    }
	    p1=pix[yw+vmin[x]];
	    p2=pix[yw+vmax[x]];

	    rsum+=((p1 & 0xff0000)-(p2 & 0xff0000))>>16;
	    gsum+=((p1 & 0x00ff00)-(p2 & 0x00ff00))>>8;
	    bsum+= (p1 & 0x0000ff)-(p2 & 0x0000ff);
	    yi++;
	  }
	  yw+=w;
	}

	for (x=0;x<w;x++){
	  rsum=gsum=bsum=0;
	  yp=-radius*w;
	  for(i=-radius;i<=radius;i++){
	    yi=max(0,yp)+x;
	    rsum+=r[yi];
	    gsum+=g[yi];
	    bsum+=b[yi];
	    yp+=w;
	  }
	  yi=x;
	  for (y=0;y<h;y++){
	    pix[yi]=0xff000000 | (dv[rsum]<<16) | (dv[gsum]<<8) | dv[bsum];
	    if(x==0){
	      vmin[y]=min(y+radius+1,hm)*w;
	      vmax[y]=max(y-radius,0)*w;
	    }
	    p1=x+vmin[y];
	    p2=x+vmax[y];

	    rsum+=r[p1]-r[p2];
	    gsum+=g[p1]-g[p2];
	    bsum+=b[p1]-b[p2];

	    yi+=w;
	  }
	}

	}

}


class MyCanvas extends ControlWindowCanvas {
	  public void draw(PApplet theApplet) {
	    theApplet.image(SwarmHunger.pg,SwarmHunger.frameX,SwarmHunger.frameY);
	    theApplet.stroke(175);
	    theApplet.strokeWeight(1);
	    theApplet.noFill();
	    theApplet.rect(SwarmHunger.frameX-1, SwarmHunger.frameY-1, SwarmHunger.frameWidth+2, SwarmHunger.frameHeight+2);
	  }
}
